/**
 * @file   cfg.c
 * @author Ondrej Jerabek
 */

#include "config.h"
#include "cfg.h"
#include "main.h"
#include "ARM_Ethernet_Reader.h"
#include "modules.h"
#include "nanotron.h"

#include <coremem.h>
#include <stddef.h>
#include <eeprom.h>
#include <eepromcrc.h>
#include <macros.h>
#include <stdio.h>
#include <sx1211config.h>
#include <multisx1211.h>
#include <multisx1211_packet.h>
#include <config-default.h>		// nanotron default config

// Functions ====================================
/**
 * @brief   Fix version of configuration specified by parameter cfg_type
 * @param   header:		Pointer to header of configuration
 * @param   cfg_type:	Type of configuration which will be fixed
 * @param   cfg_ver:	Version of configuration
 * @return  Nothing
 */
void fix_version(STRUCT_HEADER *header, enum e_cfg_type cfg_type, unsigned char cfg_ver)
{
	header->struct_type = cfg_type;
	header->device_type = DEVICE_TYPE;
	header->struct_ver = cfg_ver;
}
// --------------------------------------------------------------------------

/**
 * @brief   Load default device configuration
 * @return  Nothing
 */
void device_load_cfg_defaults(void)
{
	CLEAR(cfg);
	cfg.module1_on = true;
	cfg.module2_on = true;
	cfg.route_from_wireless1 = false;
	cfg.route_from_wireless2 = false;
	cfg.usb_enabled = false;
	cfg.nanotron_clk_en = false;
}
// --------------------------------------------------------------------------

/**
 * @brief   Load default Semtech module configuration
 * @param[out]	p_cfg:	Pointer to Semtech configuration
 * @return  Nothing
 */
void sx_load_cfg_defaults(EthernetReader_SX_Cfg *p_cfg)
{
	CLEAR(*p_cfg);
	p_cfg->rf_frequency_1 = RF_FREQUENCY_1;
	p_cfg->rf_frequency_2 = RF_FREQUENCY_2;
	p_cfg->rf_power = POUT_13DBM;
	p_cfg->if_gain = IF_GAIN_0DB;
	p_cfg->vco_trim = VCO_TRIM;
	p_cfg->clk_out = true;
	p_cfg->listen_on_chn2 = false;
}
// --------------------------------------------------------------------------

/**
 * @brief   Load default Nanotron module configuration
 * @param[out]	p_cfg:	Pointer to Nanotron configuration
 * @return  Nothing
 */
void ntr_load_cfg_defaults(EthernetReader_NTR_Cfg *p_cfg)
{
	CLEAR(*p_cfg);

	p_cfg->nano_pwr = 0x3F;					// Max nanotron PWR
	p_cfg->mode = CONFIG_DEFAULT_MODE;	// Default configuration of nanotron
}
// --------------------------------------------------------------------------

/**
 * @brief   Load default Decawave module configuration
 * @param[out]	p_cfg:	Pointer to Decawave configuration
 * @return  Nothing
 */
void dw_load_cfg_defaults(EthernetReader_DW_Cfg *p_cfg)
{
	CLEAR(*p_cfg);
	// TODO: Add DW configuration defaults
}
// --------------------------------------------------------------------------

/**
 * @brief   Load device configuration from EEPROM memory and fill device info string
 * @return  Nothing
 */
void device_load_configuration(void)
{
	STRUCT_HEADER header;
	bool cfg_loaded = false;

	snprintf((char *)info, INFO_SIZE, "ARFID Ethernet reader HW " VERSION_FORMAT_SHORT " FW " VERSION_FORMAT, VERSION_FORMAT_SHORT_ARGS(core.HwVer), VERSION_FORMAT_ARGS(core.FwVer));

	eeprom_read_block(&header, CFG_EEPROM_ADDR, sizeof(STRUCT_HEADER));

	device_load_cfg_defaults();
	if (header.device_type == DEVICE_TYPE)
	{
		if (header.struct_ver == CFG_VER_DEVICE)
			cfg_loaded = eeprom_read_block_crc(&cfg, CFG_EEPROM_ADDR, sizeof(cfg));
	}

	if (!cfg_loaded)
		device_load_cfg_defaults();

	st.header.device_type = core.DeviceType;
	st.header.struct_ver = ST_VER_DEVICE;
	st.uptime = 0;
}
// --------------------------------------------------------------------------

/**
 * @brief   Load Semtech module configuration from EEPROM memory
 * @param[out]	p_cfg:			Pointer to Semtech configuration
 * @param		addr_eeprom:	EEPROM address where the configuration is located
 * @return  Nothing
 */
void sx_load_configuration(EthernetReader_SX_Cfg *p_cfg, unsigned int addr_eeprom)
{
	STRUCT_HEADER header;
	bool cfg_loaded = false;

	eeprom_read_block(&header, addr_eeprom, sizeof(STRUCT_HEADER));

	sx_load_cfg_defaults(p_cfg);
	if (header.device_type == DEVICE_TYPE)
	{
		if (header.struct_ver == CFG_VER_SX)
			cfg_loaded = eeprom_read_block_crc(p_cfg, addr_eeprom, sizeof(EthernetReader_SX_Cfg));
	}

	if (!cfg_loaded)
		sx_load_cfg_defaults(p_cfg);

	// TODO: Load VCO configuration from eeprom on SX module
}
// --------------------------------------------------------------------------

/**
 * @brief   Load Nanotron module configuration from EEPROM memory
 * @param[out]	p_cfg:			Pointer to Nanotron configuration
 * @param		addr_eeprom:	EEPROM address where the configuration is located
 * @return  Nothing
 */
void ntr_load_configuration(EthernetReader_NTR_Cfg *p_cfg, unsigned int addr_eeprom)
{
	STRUCT_HEADER header;
	bool cfg_loaded = false;

	eeprom_read_block(&header, addr_eeprom, sizeof(STRUCT_HEADER));

	ntr_load_cfg_defaults(p_cfg);
	if (header.device_type == DEVICE_TYPE)
	{
		if (header.struct_ver == CFG_VER_NTR)
			cfg_loaded = eeprom_read_block_crc(p_cfg, addr_eeprom, sizeof(EthernetReader_NTR_Cfg));
	}

	if (!cfg_loaded)
		ntr_load_cfg_defaults(p_cfg);
}
// --------------------------------------------------------------------------

/**
 * @brief   Load Decawave module configuration from EEPROM memory
 * @param[out]	p_cfg:			Pointer to Decawave configuration
 * @param		addr_eeprom:	EEPROM address where the configuration is located
 * @return  Nothing
 */
void dw_load_configuration(EthernetReader_DW_Cfg *p_cfg, unsigned int addr_eeprom)
{
	STRUCT_HEADER header;
	bool cfg_loaded = false;

	eeprom_read_block(&header, addr_eeprom, sizeof(STRUCT_HEADER));

	dw_load_cfg_defaults(p_cfg);
	if (header.device_type == DEVICE_TYPE)
	{
		if (header.struct_ver == CFG_VER_DW)
			cfg_loaded = eeprom_read_block_crc(p_cfg, addr_eeprom, sizeof(EthernetReader_DW_Cfg));
	}

	if (!cfg_loaded)
		dw_load_cfg_defaults(p_cfg);
}
// --------------------------------------------------------------------------

/**
 * @brief   Update active channel configuration on Semtech module
 * @param	module:	Module index
 * @return  Nothing
 */
void change_channel(enum e_module module)
{
	switch(module)
	{
		case MODULE_1:
			if (module_cfg[module].sx.listen_on_chn2)
				sx_change_rps_1(RPS_SELECT_2);
			else
				sx_change_rps_1(RPS_SELECT_1);
			wireless_start_receive_1();
			break;
		case MODULE_2:
			if (module_cfg[module].sx.listen_on_chn2)
				sx_change_rps_2(RPS_SELECT_2);
			else
				sx_change_rps_2(RPS_SELECT_1);
			wireless_start_receive_2();
			break;
	}
}
// --------------------------------------------------------------------------

/**
 * @brief   Load configuration specified by parameter cfg_type
 * @param   cfg_type:	Type of configuration which will be loaded
 * @return  Nothing
 */
void load_configuration(enum e_cfg_type cfg_type)
{
	unsigned int address = CFG_MODULE_ADDR;
	unsigned char module = 0;

	switch (cfg_type)
	{
		case CFG_DEVICE:
			device_load_configuration();
			fix_version(&cfg.header, CFG_DEVICE, CFG_VER_DEVICE);
			break;

		case CFG_MODULE2_SX:
			address += CFG_MODULE2_OFFSET;
			module = 1;
		case CFG_MODULE1_SX:
			address += CFG_SX_OFFSET;
			sx_load_configuration(&module_cfg[module].sx, address);
			fix_version(&module_cfg[module].sx.header, cfg_type, CFG_VER_SX);
			break;

		case CFG_MODULE2_NTR:
			address += CFG_MODULE2_OFFSET;
			module = 1;
		case CFG_MODULE1_NTR:
			address += CFG_NTR_OFFSET;
			ntr_load_configuration(&module_cfg[module].ntr, address);
			fix_version(&module_cfg[module].ntr.header, cfg_type, CFG_VER_NTR);
			break;

		case CFG_MODULE2_DW:
			address += CFG_MODULE2_OFFSET;
			module = 1;
		case CFG_MODULE1_DW:
			address += CFG_DW_OFFSET;
			dw_load_configuration(&module_cfg[module].dw, address);
			fix_version(&module_cfg[module].dw.header, cfg_type, CFG_VER_DW);
			break;

		case CFG_ALL:
			load_configuration(CFG_DEVICE);
			load_configuration(CFG_MODULE1_SX);
			load_configuration(CFG_MODULE2_SX);
			load_configuration(CFG_MODULE1_NTR);
			load_configuration(CFG_MODULE2_NTR);
			load_configuration(CFG_MODULE1_DW);
			load_configuration(CFG_MODULE2_DW);
			break;
		default: return;
	}

	flags.update_cfg = true;
}
// --------------------------------------------------------------------------

/**
 * @brief   Save configuration specified by parameter cfg_type
 * @param   cfg_type:	Type of configuration which will be saved
 * @return  Nothing
 */
void save_configuration(enum e_cfg_type cfg_type)
{
	unsigned int address = CFG_MODULE_ADDR;
	unsigned char module = 0;

	switch (cfg_type)
	{
		case CFG_DEVICE:
			fix_version(&cfg.header, CFG_DEVICE, CFG_VER_DEVICE);
			eeprom_write_block_crc(&cfg, CFG_EEPROM_ADDR, sizeof(cfg));
			break;

		case CFG_MODULE2_SX:
			address += CFG_MODULE2_OFFSET;
			module = 1;
		case CFG_MODULE1_SX:
			address += CFG_SX_OFFSET;
			fix_version(&module_cfg[module].sx.header, cfg_type, CFG_VER_SX);
			eeprom_write_block_crc(&module_cfg[module].sx, address, sizeof(EthernetReader_SX_Cfg));
			break;

		case CFG_MODULE2_NTR:
			address += CFG_MODULE2_OFFSET;
			module = 1;
		case CFG_MODULE1_NTR:
			address += CFG_NTR_OFFSET;
			fix_version(&module_cfg[module].ntr.header, cfg_type, CFG_VER_NTR);
			eeprom_write_block_crc(&module_cfg[module].ntr, address, sizeof(EthernetReader_NTR_Cfg));
			break;

		case CFG_MODULE2_DW:
			address += CFG_MODULE2_OFFSET;
			module = 1;
		case CFG_MODULE1_DW:
			address += CFG_DW_OFFSET;
			fix_version(&module_cfg[module].dw.header, cfg_type, CFG_VER_DW);
			eeprom_write_block_crc(&module_cfg[module].dw, address, sizeof(EthernetReader_DW_Cfg));
			break;

		case CFG_ALL:
			save_configuration(CFG_DEVICE);
			save_configuration(CFG_MODULE1_SX);
			save_configuration(CFG_MODULE2_SX);
			save_configuration(CFG_MODULE1_NTR);
			save_configuration(CFG_MODULE2_NTR);
			save_configuration(CFG_MODULE1_DW);
			save_configuration(CFG_MODULE2_DW);
			break;
		default: return;
	}
}
// --------------------------------------------------------------------------

/**
 * @brief   Update all device configurations
 * @return  Nothing
 */
void update_configuration()
{
	// Module 1 update
	switch(modules.slot1_type)
	{
		case MODULE_SX:
			sx_change_mode_1(SXM_FS);
			sx_write_reg_1(R07_P1, P_CNT(module_cfg[MODULE_1].sx.rf_frequency_1));
			sx_write_reg_1(R08_S1, S_CNT(module_cfg[MODULE_1].sx.rf_frequency_1));
			sx_write_reg_1(R10_P2, P_CNT(module_cfg[MODULE_1].sx.rf_frequency_2));
			sx_write_reg_1(R11_S2, S_CNT(module_cfg[MODULE_1].sx.rf_frequency_2));
			vco_trim_1 = module_cfg[MODULE_1].sx.vco_trim;
			sx_write_reg_1(R01_MCPARAM2, (sx_config_1[R01_MCPARAM2] & 0xFC) | (module_cfg[MODULE_1].sx.if_gain));
			sx_write_reg_1(R26_TXPARAM,  (sx_config_1[R26_TXPARAM] & 0xF1)  | (module_cfg[MODULE_1].sx.rf_power));
			sx_write_reg_1(R27_OSCPARAM, (sx_config_1[R27_OSCPARAM] & 0x7F) | (module_cfg[MODULE_1].sx.clk_out << 7 ));
			change_channel(MODULE_1);
			wireless_start_receive_1();
			break;

		case MODULE_NTR:
			nanotron_init(MODULE_1);		// Reinit with new mode
			nanotron_set_tx_pwr(nanotron_get_ptr(MODULE_1), module_cfg[MODULE_1].ntr.nano_pwr);
			break;

		case MODULE_DW:
			// TODO: Set new configuration to DW module
			break;
		default:
			break;
	}

	// Module 2 update
	switch(modules.slot2_type)
	{
		case MODULE_SX:
			sx_change_mode_2(SXM_FS);
			sx_write_reg_2(R07_P1, P_CNT(module_cfg[MODULE_2].sx.rf_frequency_1));
			sx_write_reg_2(R08_S1, S_CNT(module_cfg[MODULE_2].sx.rf_frequency_1));
			sx_write_reg_2(R10_P2, P_CNT(module_cfg[MODULE_2].sx.rf_frequency_2));
			sx_write_reg_2(R11_S2, S_CNT(module_cfg[MODULE_2].sx.rf_frequency_2));
			vco_trim_2 = module_cfg[MODULE_2].sx.vco_trim;
			sx_write_reg_2(R01_MCPARAM2, (sx_config_2[R01_MCPARAM2] & 0xFC) | (module_cfg[MODULE_2].sx.if_gain));
			sx_write_reg_1(R26_TXPARAM,  (sx_config_1[R26_TXPARAM] & 0xF1)  | (module_cfg[MODULE_2].sx.rf_power));
			sx_write_reg_1(R27_OSCPARAM, (sx_config_1[R27_OSCPARAM] & 0x7F) | (module_cfg[MODULE_2].sx.clk_out << 7 ));
			change_channel(MODULE_2);
			wireless_start_receive_2();
			break;
		case MODULE_NTR:
			nanotron_init(MODULE_2);		// Reinit with new mode
			nanotron_set_tx_pwr(nanotron_get_ptr(MODULE_2), module_cfg[MODULE_2].ntr.nano_pwr);
			break;

		case MODULE_DW:
			// TODO: Set new configuration to DW module
			break;

		default:
			break;
	}

	// Device update
	SystemClkOut_init(cfg.nanotron_clk_en);
}
// --------------------------------------------------------------------------

