/**
 * @file   nanotron.c
 * @author Ondrej Jerabek
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "nanotron.h"
#include "config.h"
#include <ntrxtypes.h>
#include <ntrxutil.h>
#include <ntrxranging.h>
#include "../lib/multinanotron/include/phy.h"
#include <nnspi.h>
#include <hwclock.h>
#include "main.h"
#include "modules.h"

#include "lpc_timer.h"
#include <is_rqstat.h>

#include <type.h>
#include <25XX640.h>
#include <interrupt.h>
#include <coremem.h>
#include <uart_packet.h>
#include <ntrx_packet.h>
#include <rtcsync.h>
#include <extcomm.h>
#include <wdt.h>

/*#include "ranging.h"
#include "synchronization.h"
#include "anchor.h"
#include "timesync.h"
#include "faultlog.h"*/

extern void HardFault_Handler(void);
extern void ErrorHandler(int16_t err);

MsgT downMsg;

// --------------------------------------------------------------------------
NanoGlobCTX nanotrons[NTRX_COUNT];
TOA_DATA toa[NTRX_COUNT];

// --------------------------------------------------------------------------

#define NTRX_SELECT(i)				NTRX_##i##_CS(0)
#define NTRX_UNSELECT(i)			NTRX_##i##_CS(1)
static void NTRXSelect(void *self, NanoGlobCTX *ngX, bool selected) {
	switch ((uint32_t)self) {
	case 0:
		if (selected)
			NTRX_SELECT(1);
		else
			NTRX_UNSELECT(1);
		break;
	case 1:
		if (selected)
			NTRX_SELECT(2);
		else
			NTRX_UNSELECT(2);
		break;
	default:
		ErrorHandler(2);
		break;
	}
}
// --------------------------------------------------------------------------

/**
 * @brief 	Reading ToA registers from nanoLOC chip
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return 	Nothing
 * @note	Function called from nanotron library
 *
 */
static void ReadToaRegisters(void * self, NanoGlobCTX * ngX) {
	uint32_t index = (uint32_t)self;
	TOA_DATA * _toa = &toa[index];
	//_toa->BaseTimer = ntrx_irq_stamp[index];

	/* Read Tx Ranging Registers */
	NTRXSPIReadByte(&nanotrons[index], NA_ToaOffsetMeanDataValid_O,
			&_toa->ToaOffsetMeanDataValid);
	_toa->ToaOffsetMeanDataValid &= (0x01 << NA_ToaOffsetMeanDataValid_B);
	_toa->ToaOffsetMeanDataValid = (uint8_t) (_toa->ToaOffsetMeanDataValid
			>> NA_ToaOffsetMeanDataValid_B);

	NTRXSPIReadByte(ngX, NA_PhaseOffsetData_O, &_toa->PhaseOffsetData);
	_toa->PhaseOffsetData = (_toa->PhaseOffsetData >> NA_PhaseOffsetData_LSB)
			& 0x07;

	NTRXSPIRead(ngX, NA_ToaOffsetMeanData_O,
			(uint8_t *) &_toa->ToaOffsetMeanData, 2);
	_toa->ToaOffsetMeanData &= 0x1fff;

#if 0
	// vycteni casu odpovedi
	NTRXSPIReadByte (NA_ToaOffsetMeanAckValid_O, &toa.ToaOffsetMeanAckValid);
	toa.ToaOffsetMeanAckValid &= (0x01 << NA_ToaOffsetMeanAckValid_B);
	toa.ToaOffsetMeanAckValid = (uint8_t)(toa.ToaOffsetMeanAckValid >> NA_ToaOffsetMeanAckValid_B);

	/*	if(ToaOffsetMeanAckValid != 1)
	 {
	 printf("ERROR ReadRegisterAck\n");
	 }
	 else*/
	{
		NTRXSPIReadByte(NA_PhaseOffsetAck_O, &toa.PhaseOffsetAck);
		toa.PhaseOffsetAck = (toa.PhaseOffsetAck >> NA_PhaseOffsetAck_LSB) & 0x07;

		NTRXSPIRead(NA_ToaOffsetMeanAck_O, (uint8_t *) &toa.ToaOffsetMeanAck, 2);
		toa.ToaOffsetMeanAck &= 0x1fff;
	}

	NTRXSPIRead(NA_TxRespTime_O, (uint8_t *) &toa.TxRespTime, sizeof(toa.TxRespTime));
#endif
}
// --------------------------------------------------------------------------

/**
 * @brief 	Error message handler
 * @param	err: Error number
 * @return 	Nothing
 */
void ErrorHandler(int16_t err) {
	const char * err_str = "Unknown";
	bool reset = true;
	switch (err) {
		case 1: err_str = "Payload overflow"; break;
		case 2: err_str = "Invalid SPI index"; break;
		case 506: err_str = "Initialization failure"; reset = false; break;
		default: break;
	}

	//notify_error_condition(ERR_READER_NTRX, err, "Critical nanotron error #%d (%s)!", err, err_str);
	if (reset) {
		led.err = 2000;
		while(led.err) WDT_RESET();
		NVIC_SystemReset();
	}
}
// --------------------------------------------------------------------------

/**
 * @brief 	Process incoming messages.
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @param 	msg:	This is the message pointer
 * @return	Nothing
 * @note	Function called from nanotron library
 */
static void APLCallback(void *self, NanoGlobCTX * ngX, MsgT *msg) {
	switch (msg->prim) {
	case PD_TIMEB_CONFIRM:
	case PD_DATA_CONFIRM: /* process message send confirmation here */
#ifndef NTRX_RANGER_TEST
		switch (msg->status) {
		case PHY_SUCCESS:
			//anchor_handle_tx(ngX);
			break;
		}
#endif
		break;

	case PD_DATA_INDICATION:/* process incomming messages here */
#ifndef NTRX_RANGER_TEST
		if (ngX->index == 0) led.mod1_blink = true;
		else if (ngX->index == 1) led.mod2_blink = true;
		ntrx_handle_packet(ngX, msg);
#endif
//			printf("DI [%02u] %8lu %s\n", msg->len, t, msg->data);
		break;

	case PD_BRDC_INDICATION:
#ifndef NTRX_RANGER_TEST
		if (ngX->index == 0) led.mod1_blink = true;
		else if (ngX->index == 1) led.mod2_blink = true;
		ntrx_handle_packet(ngX, msg);
#endif
//			printf("DC [%02u] %lu\n", msg->len, t);
//			printf("BC %u\n", ntrx_irq_stamp);
		break;

	case PD_RANGING_CONFIRM:
	case PD_RANGING_INDICATION:
		// TODO: Write ranging callback
		//ranging_callback(ngX, msg);
		break;

	case PD_TIMEB_INDICATION:
		// TODO: Write Time broadcast processing
		//process_timeb_packet(ngX, msg);
		led.sys = 20;
		break;

	default:
		break;
	}
}
// --------------------------------------------------------------------------

/**
 * @brief   Initialize one nanoLOC module
 * @param	index:	NanoLOC chip index
 * @return  Initialization result. True if successful
 */
bool nanotron_init(unsigned int index) {
	NanoGlobCTX * ngX = &nanotrons[index];
	uint8_t srcAddress[] = { 0x02, 0, 0, 0, 0, 0 };	// locally administered address template
	uint8_t * ntrxShadowReg;
	struct APLCallback_t aplCallback = (struct APLCallback_t){
		.cb   = APLCallback,
		.self = (void *)index
	};
	struct NtrxSpiSelect spiSelect = (struct NtrxSpiSelect){
		.cb   = NTRXSelect,
		.self = (void *)index
	};

	struct NtrxCallback ntrxCallback = (struct NtrxCallback){
		.cb   = ReadToaRegisters,
		.self = (void *)index
	};

	S_NTRX_util_init(&ngX->util);
	S_NTRX_util_SetChannel_init(&ngX->channel);
	S_NTRX_ranging_init(&ngX->ranging);
	S_NanotronPhy_Init(&ngX->physical, aplCallback, spiSelect, ntrxCallback);
	ngX->index = (uint8_t) index;
#ifdef CONFIG_NTRX_PROMISCUOUS
	ngX->physical.MsgFilter = MsgFilterCallback;
#endif

	ntrx_addr2mac(srcAddress, core.MyAddr, ngX->index);

	/*printf("src: %02X:%02X:%02X:%02X:%02X:%02X\n", srcAddress[0], srcAddress[1],
			srcAddress[2], srcAddress[3], srcAddress[4], srcAddress[5]);*/

	/*
	 * Initialize the ntrx driver
	 */
	NTRXSetDefaultMode(ngX, module_cfg[index].ntr.mode);
	if (NTRXInit(ngX) == FALSE)
		return false;

	/* set MAC address in nanoLOC receiver */
	downMsg.prim = PLME_SET_REQUEST;
	downMsg.attribute = PHY_MAC_ADDRESS1;
	memcpy(downMsg.data, srcAddress, 6);
	PLMESap(ngX, &downMsg);

	/* switch on receiver */
	downMsg.prim = PLME_SET_REQUEST;
	downMsg.attribute = PHY_RX_CMD;
	downMsg.value = PHY_RX_ON;
	PLMESap(ngX, &downMsg);
#if 0
	downMsg.prim = PHY_ADDR_MATCHING;
	downMsg.value = 0;
	PLMESap( &downMsg );
#endif
#if 0
	downMsg.prim = PLME_SET_REQUEST;
	downMsg.attribute = PHY_RTC_AUTOMODE;
	downMsg.value = true;
	PLMESap(&downMsg);
#endif

	ntrxShadowReg = ngX->util.ntrxShadowReg;

#if CONFIG_DEFAULT_MODE == CONFIG_NTRX_80MHZ_1000NS
	ntrxShadowReg[NA_RtcTimeBRxAdj_O] = NA_MacRtcTimeBRxAdj1M2Ary_IC_C;
	ntrxShadowReg[NA_RtcTimeBTxAdj_O] = NA_MacRtcTimeBTxAdj1M2Ary_IC_C;
#elif  CONFIG_DEFAULT_MODE == CONFIG_NTRX_80MHZ_2000NS
	ntrxShadowReg[NA_RtcTimeBRxAdj_O] = NA_MacRtcTimeBRxAdj500k2Ary_IC_C;
	ntrxShadowReg[NA_RtcTimeBTxAdj_O] = NA_MacRtcTimeBTxAdj500k2Ary_IC_C;
#elif  CONFIG_DEFAULT_MODE == CONFIG_NTRX_80MHZ_4000NS || CONFIG_DEFAULT_MODE == CONFIG_NTRX_22MHZ_4000NS
	ntrxShadowReg[NA_RtcTimeBRxAdj_O] = NA_MacRtcTimeBRxAdj250k2Ary_IC_C;
	ntrxShadowReg[NA_RtcTimeBTxAdj_O] = NA_MacRtcTimeBTxAdj250k2Ary_IC_C;
#endif
	NTRXSPIWriteByte(ngX, NA_RtcTimeBRxAdj_O,
			ntrxShadowReg[NA_RtcTimeBRxAdj_O]);
	NTRXSPIWriteByte(ngX, NA_RtcTimeBTxAdj_O,
			ntrxShadowReg[NA_RtcTimeBTxAdj_O]);

#ifdef SUPPORT_RANGING_TYPE_2W_PP
	RangingRegisterCallback(Permission, GetMemForPHY);
#endif

	NTRXAllCalibration(ngX);
	return true;
}
// --------------------------------------------------------------------------

/**
 * @brief   Initialize all nanoLOC modules
 * @return  Initialization result. True if successful
 */
bool nanotron_init_all(void) {
	register uint32_t index;
	register bool init = true;

	for (index = 0; index < NTRX_COUNT; index++)
		init = nanotron_init(index) && init;

	return init;
}
// --------------------------------------------------------------------------

/**
 * @brief   Calibrate all nanoLOC modules
 * @return  Calibration result. True if successful
 */
bool nanotron_calibrate_all(void) {
	return NTRXAllCalibrationMulti(&nanotrons[0], NTRX_COUNT);
}
// --------------------------------------------------------------------------

/**
 * @brief   Calibrate one NanoLOC module
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  Calibration result. True if successful
 */
bool nanotron_calibrate(NanoGlobCTX * ngX) {
	return NTRXAllCalibrationMulti(ngX, 1);
}
// --------------------------------------------------------------------------

/**
 * @brief   Enable NanoLOC receiver
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  Nothing
 */
void nanotron_rx_enable(NanoGlobCTX * ngX, bool enabled) {
	downMsg.prim = PLME_SET_REQUEST;
	downMsg.attribute = PHY_RX_CMD;
	downMsg.value = enabled ? PHY_RX_ON : PHY_TRX_OFF;
	PLMESap(ngX, &downMsg);
}
// --------------------------------------------------------------------------

/**
 * @brief   Sending RTC broadcast packet through nanoLOC module
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  Nothing
 */
void nanotron_broadcast_rtc(NanoGlobCTX * ngX) {
	memset(downMsg.addr, 0, 6);
	downMsg.prim = PD_TIMEB_REQUEST;
	downMsg.len = 1;
	downMsg.data[0] = 1;
	SendMsg(ngX, &downMsg);
}
// --------------------------------------------------------------------------

/**
 * @brief   Handling nanoLOC communication
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  Nothing
 */
void nanotron_task(NanoGlobCTX * ngX) {
#ifdef CONFIG_NTRX_IRQ
	if (nnIrq != 0)
	PHYPoll();
#ifdef CONFIG_TRAFFIC_LED
	UpdateLEDs();
#endif /* CONFIG_TRAFFIC_LED */
	//		KeyUpdate();
#ifdef CONFIG_NTRX_AUTO_RECALIB
	PHYAutoCalibration();
#endif /* CONFIG_NTRX_AUTO_RECALIB */
#else
	if (ngX->physical.phyPIB.pwrDown == FALSE)
		PHYPoll(ngX);
#endif
	ntrx_pck_task(ngX);
}
// --------------------------------------------------------------------------

/**
 * @brief   Enable receiving all packets (Disable address checking)
 * @param	index:		NanoLOC chip index
 * @param	enabled:	Enable/disable receiving all packets
 * @return  Nothing
 */
void nanotron_enable_promiscuous(unsigned int index, bool enabled) {
	downMsg.prim = PLME_SET_REQUEST;
	downMsg.attribute = PHY_ADDR_MATCHING;
	downMsg.value = enabled ? 0 : 1;
	PLMESap(nanotron_get_ptr(index), &downMsg);
}
// --------------------------------------------------------------------------

/**
 * @brief   Read ToA (Time of Arrival) data structure
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  Pointer to TOA data structure
 */
TOA_DATA * nanotron_get_toa(NanoGlobCTX * ngX) {
	return &toa[ngX->index];
}
// --------------------------------------------------------------------------

/**
 * @brief   Translate nanoLOC index to pointer to NanoLOC work & configuration structure
 * @param	index:	NanoLOC chip index
 * @return  Pointer to NanoLOC work & configuration structure
 */
inline NanoGlobCTX * nanotron_get_ptr(unsigned int index) {
	return &nanotrons[index];
}
// --------------------------------------------------------------------------

/**
 * @brief   Send timesync packet for NanoLOC modules
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  Result of packet sending
 * @warning	Not implemented yet. For testing purpose, sends only device status
 */
enum e_send_receive_status send_timesync_packet(NanoGlobCTX * ngX)
{
	TPacket pck;
	pck.Addr_Src = core.MyAddr;
	pck.Addr_Dest = BROADCAST;
	pck.Rq_St = STAT_IS_READER_BROADCAST;
	pck.Data_Len = sizeof(EthernetReader_State);
	pck.Buffer = (unsigned char *) &st;
	pck.Pck_Type = ROUTING(RTD_BOTH);

	ntrx_pck_send(ngX, &pck);

	switch (downMsg.status)
	{
		case PHY_SUCCESS :
			return SRS_OK;				// packet sent, ACK received (for unicast packet)
		case PHY_NO_ACK :
			return SRS_NO_ANSWER;		// packet sent, NO ACK
		case PHY_BUSY_TX :
			return SRS_LINK_BUSY;		// BUSY - TX already in progress
		case PHY_TX_TIMESLOT_TIMEOUT :
		default :
			return SRS_TRANSMIT_ERROR;	// other error
	}
}
// --------------------------------------------------------------------------

/**
 * @brief   Sending Beacon through nanotron module
 * @return  Nothing
 */
void nanotron_beacon(void) {
	static int beacon_counter = 0;
	static int next_calib = 0;
	static bool need_recalib = true;
	static bool ntrx_ready_1, ntrx_ready_2;

	NanoGlobCTX *ngXL = nanotron_get_ptr(0);
	NanoGlobCTX *ngXR = nanotron_get_ptr(1);

#ifndef NTRX_BEACON_CONTINUAL
	if (!tmr.timesync_broadcast)
#endif

	{
		if (need_recalib || next_calib <= 0) {
			next_calib = (rand() & 0x03) * 1000;

			// Calibrate NTRX always!
			if (modules.slot1_type == MODULE_NTR && modules.slot2_type == MODULE_NTR)
				need_recalib = !nanotron_calibrate_all();
			else if (modules.slot1_type == MODULE_NTR)
				need_recalib = !nanotron_calibrate(nanotron_get_ptr(0));
			else if (modules.slot2_type == MODULE_NTR)
				need_recalib = !nanotron_calibrate(nanotron_get_ptr(1));

			ntrx_ready_1 = modules.slot1_type == MODULE_NTR && ngXL->util.calib_rx_ok && ngXL->util.calib_tx_ok;
			ntrx_ready_2 = modules.slot2_type == MODULE_NTR  && ngXR->util.calib_rx_ok && ngXR->util.calib_tx_ok;
		}

		// odesila cas na broadcast pres nanoLOC
#ifdef NTRX_BEACON_CONTINUAL
		if (!tmr.timesync_broadcast)
#endif
		{

			tmr.timesync_broadcast = TIMESYNC_BROADCAST_TIMEOUT / 4 * 3
					+ rand() % (TIMESYNC_BROADCAST_TIMEOUT / 4);
			next_calib -= tmr.timesync_broadcast;
		}

		if (ntrx_ready_1 && ntrx_ready_2)
		{
			if (beacon_counter & 1)
				led.mod2_blink = 1;
			else
				led.mod1_blink = 1;
			send_timesync_packet(nanotron_get_ptr(beacon_counter++ & 1));
		}
		else {
			if (ntrx_ready_1)
			{
				send_timesync_packet(nanotron_get_ptr(0));
				led.mod1_blink = 1;
			}
			if (ntrx_ready_2)
			{
				send_timesync_packet(nanotron_get_ptr(1));
				led.mod2_blink = 1;
			}
		}
	}
}
// --------------------------------------------------------------------------

/**
 * @brief   Set nanoLOC chip to power down mode
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @return  @c false.
 */
bool nanotron_pwr_down(NanoGlobCTX * ngX)
{
	downMsg.value = 0;
	downMsg.attribute = PHY_PWR_DOWN_MODE;
	downMsg.prim = PLME_SET_REQUEST;
	PLMESap(ngX, &downMsg);

	return false;
}
// --------------------------------------------------------------------------

/**
 * @brief   Set nanoLOC chip transmitting power
 * @param	ngX:	Pointer to NanoLOC work & configuration structure
 * @param	pwr:	NanoLOC transmit power. Valid value 0 - 63
 * @return  @c true if set succesfully
 */
bool nanotron_set_tx_pwr(NanoGlobCTX * ngX, unsigned char pwr)
{
	if (pwr < 0 || pwr > 0x3F)
		return false;

	downMsg.value = pwr;
	downMsg.attribute = PHY_TX_POWER;
	downMsg.prim = PLME_SET_REQUEST;
	PLMESap(ngX, &downMsg);

	return false;
}
