/**
 * @file   power_ctrl.c
 * @author Ondrej Jerabek
 */

#include "power_ctrl.h"
#include "pindef.h"
#include "main.h"

#include <adc_17xx_40xx.h>
#include <wdt.h>

// --------------------------------------------------------------------------
/**
 * @brief  	Averaging structure
 */
typedef struct
{
	unsigned long sum;			/**< Sum of input values	*/
	unsigned char samples;		/**< Number of sum samples	*/
} AvgStruct;

// --------------------------------------------------------------------------
/* Global variables */
AvgStruct avg[4];

// --------------------------------------------------------------------------
/**
 * @brief  	Averaging measured values
 * @param	avg_t:	Pointer to averaging structure
 * @param	in_val:	Input value for averaging
 * @param	result:	Pointer to variable where result be stored
 * @return  Nothing
 */
void avg_handle(AvgStruct *avg_t, unsigned short in_val, unsigned short *result)
{
	avg_t->sum += in_val;
	avg_t->samples++;
	if (avg_t->samples >= AVG_SAMPLES)
	{
		*result = avg_t->sum / avg_t->samples;
		avg_t->sum = 0;
		avg_t->samples = 0;
	}
}
// --------------------------------------------------------------------------

/**
 * @brief   Set charger power mode
 * @param	mode:	Current mode. Valid values avaible in enum e_chrg_mode
 * @return  Nothing
 */
void set_chrg_mode(enum e_chrg_mode mode)
{
	switch (mode)
	{
		case CHRG_OFF:
			CHRG_CE(1);
			CHRG_EN1(0);
			CHRG_EN2(1);
			break;

		case CHRG_100MA:		// Not recomended. Device consumes more current than 100 mA. => remaining current will be delivered from battery
			CHRG_CE(0);
			CHRG_EN1(0);
			CHRG_EN2(0);
			break;

		case CHRG_500MA:
			CHRG_CE(0);
			CHRG_EN1(1);
			CHRG_EN2(0);
			break;

		case CHRG_600MA:
			CHRG_CE(0);
			CHRG_EN1(0);
			CHRG_EN2(1);
			break;

		case CHRG_BATT_ONLY:
			CHRG_CE(0);
			CHRG_EN1(1);
			CHRG_EN2(1);
			break;

		default:
			return;
			break;
	}
	st.chrg_mode = mode;
}
// --------------------------------------------------------------------------

/**
 * @brief  	Device power off
 * @return  Nothing
 * @note	If device powered from external power, does nothing
 */
void pwr_suicide(void)
{
	unsigned char i;

	if (!CHRG_PWR_GOOD)		// powered from external power, can't do this
		return;

	for (i = 0; i < 10; i++)
	{
		WDT_RESET();
		tmr.buzz = 100;
		delay(200);
	}

	while (1)
		POWER(0);
}
// --------------------------------------------------------------------------

/**
 * @brief  	Handling ADC measurement
 * @return  Nothing
 */
void pwr_meas_vals(void)
{
	unsigned short result;
	if (Chip_ADC_ReadValue(LPC_ADC, V_BATT_CHNL, &result) == SUCCESS)
		avg_handle(&avg[0], result * DIVIDER(DIV_BATT_R1, DIV_BATT_R2), &st.batt_voltage);

	if (Chip_ADC_ReadValue(LPC_ADC, V_SUPPLY_CHNL, &result) == SUCCESS)
		avg_handle(&avg[1], result * DIVIDER(DIV_SUPPLY_R1, DIV_SUPPLY_R2), &st.supply_voltage);

	if (Chip_ADC_ReadValue(LPC_ADC, V_USB_CHNL, &result) == SUCCESS)
		avg_handle(&avg[2], result * DIVIDER(DIV_USB_R1, DIV_USB_R2), &st.usb_voltage);

	if (Chip_ADC_ReadValue(LPC_ADC, I_MEAS_CHNL, &result) == SUCCESS)
		avg_handle(&avg[3], result * VREF * 400 / (ADC_LEVELS * R_ISET), &st.chrg_current);
}
// --------------------------------------------------------------------------

/**
 * @brief  	Indication power states on LED and Buzzer
 * @return  Nothing
 */
void pwr_indication(void)
{
	// PWR LED indication
	if (!CHRG_CHARGING)			// Connected to power source, charging in progress
		led.pwr = 950;
	else if (!CHRG_PWR_GOOD && st.chrg_mode != CHRG_BATT_ONLY)	// Connected to power source, not charging / charged
		led.pwr = 1000;
	else if (st.batt_voltage < BATT_LOW_VOLTAGE)	// Powered from battery, low voltage
	{
		led.pwr = 500;
		led.err = 500;
	}
	else						// Powered from battery
		led.pwr = 50;

	// Buzzer indication
	if (st.uptime % 10 == 0)	// every 10 sec
	{
		if ((st.batt_voltage < BATT_LOW_VOLTAGE) && CHRG_CHARGING && CHRG_PWR_GOOD)
			tmr.buzz = 500;
	}
}
// --------------------------------------------------------------------------

/**
 * @brief  	Handling power control and state indication
 * @return  Nothing
 */
void pwr_handle(void)
{
	static unsigned char last_pwr_state = 1;

	pwr_indication();

	st.charging = !CHRG_CHARGING;
	st.pwr_good = !CHRG_PWR_GOOD;
	if (last_pwr_state != st.pwr_good)		// Sound after changing power state
		tmr.buzz = 1000;
	last_pwr_state = st.pwr_good;

	if (cfg.usb_enabled)	// if USB enabled
	{
		if ((st.usb_voltage > 4000 && !flags.usb_forced) || (st.usb_voltage < 3500 && flags.usb_forced))		// USB cable connected or disconnected
			NVIC_SystemReset();								// Reset device and reinitialize components
	}

	if (st.uptime % 10 == 0)	// Check voltage every 10 sec
	{
		if (st.batt_voltage < BATT_CRITICAL_VOLTAGE)
			flags.power_off = true;		// Power off device
	}
}
// --------------------------------------------------------------------------
